// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:meta/meta.dart';
import 'package:wyatt_form_bloc/src/enums/enums.dart';

/// {@template form_input}
/// A [FormInput] represents the value of a single form input field.
/// It contains information about the [FormInputStatus], [value], as well
/// as validation status.
///
/// [FormInput] should be extended to define custom [FormInput] instances.
///
/// ```dart
/// enum FirstNameError { empty }
/// class FirstName extends FormInput<String, FirstNameError> {
///   const FirstName.pure({String value = ''}) : super.pure(value);
///   const FirstName.dirty({String value = ''}) : super.dirty(value);
///
///   @override
///   FirstNameError? validator(String value) {
///     return value.isEmpty ? FirstNameError.empty : null;
///   }
/// }
/// ```
/// {@endtemplate}
@immutable
abstract class FormInput<T, E> {
  const FormInput._(this.value, [this.pure = true]);

  /// Constructor which create a `pure` [FormInput] with a given value.
  const FormInput.pure(T value) : this._(value);

  /// Constructor which create a `dirty` [FormInput] with a given value.
  const FormInput.dirty(T value) : this._(value, false);

  /// The value of the given [FormInput].
  /// For example, if you have a `FormInput` for `FirstName`,
  /// the value could be 'Joe'.
  final T value;

  /// If the [FormInput] is pure (has been touched/modified).
  /// Typically when the `FormInput` is initially created,
  /// it is created using the `FormInput.pure` constructor to
  /// signify that the user has not modified it.
  ///
  /// For subsequent changes (in response to user input), the
  /// `FormInput.dirty` constructor should be used to signify that
  /// the `FormInput` has been manipulated.
  final bool pure;

  /// The [FormInputStatus] which can be one of the following:
  /// * [FormInputStatus.pure]
  ///   - if the input has not been modified.
  /// * [FormInputStatus.invalid]
  ///   - if the input has been modified and validation failed.
  /// * [FormInputStatus.valid]
  ///   - if the input has been modified and validation succeeded.
  FormInputStatus get status => pure
      ? FormInputStatus.pure
      : valid
          ? FormInputStatus.valid
          : FormInputStatus.invalid;

  /// Returns a validation error if the [FormInput] is invalid.
  /// Returns `null` if the [FormInput] is valid.
  E? get error => validator(value);

  /// Whether the [FormInput] value is valid according to the
  /// overridden `validator`.
  ///
  /// Returns `true` if `validator` returns `null` for the
  /// current [FormInput] value and `false` otherwise.
  bool get valid => validator(value) == null;

  /// Whether the [FormInput] value is not valid.
  /// A value is invalid when the overridden `validator`
  /// returns an error (non-null value).
  bool get invalid => status == FormInputStatus.invalid;

  /// A function that must return a validation error if the provided
  /// [value] is invalid and `null` otherwise.
  E? validator(T value);

  @override
  int get hashCode => value.hashCode ^ pure.hashCode;

  @override
  bool operator ==(Object other) {
    if (other.runtimeType != runtimeType) return false;
    return other is FormInput<T, E> &&
        other.value == value &&
        other.pure == pure;
  }

  @override
  String toString() => 'FormInput<$runtimeType>(value: $value, pure: $pure)';
}
